/*	mod_kssl.c  --  Adds kerberos config directives to apache.  Needs mod_ssl.
**
**	19991109 	VRS 	Started.
**	20000512	VRS 	Added port access control via random nonce in
**						Cookies:, Accept_Language: headers.
*/

#ifndef NO_KRB5

#include <strings.h>
#include "../ssl/mod_ssl.h"
#include "mod_kssl.h"

#ifdef  KSSL_KLP
#include "ap_mitm.h"
#endif  /* KSSL_KLP */


module	kssl_module;


/*	20000512 VRS  Recast cookie as static global. 
**	It was in srvr_config, but there turned out to be multiple srvr_configs.
*/
Cookie	*cookie = NULL;		/*  Used by proxy for access control for port	*/


Cookie	*new_cookie()
{
	if (cookie)  {
		if (cookie->nonce)  free(cookie->nonce);
		free(cookie);
		}
	if ((cookie = (Cookie *) calloc(1, sizeof(Cookie))))  {
		cookie->nonce_len = 8;
		cookie->max_tries = 5;
		}

	return cookie;
}


char	*new_nonce(const char *val)
{
	if (!cookie)  return NULL;
	if (cookie->nonce)  free(cookie->nonce);

	cookie->nonce = NULL;
	if (val  &&  *val)  {
		cookie->nonce = (char *) calloc(1, strlen(val) + 1);
		strcpy(cookie->nonce, val);
		}

	return cookie->nonce;
}


/*	For all ar_scheme and ar_https table elements in kssl_srvr_cfg,
**	build a domainlist formatted as: elem1\0elem2\0...elemN\0\0.
**	Return domainlist on success, NULL on failure.
**	Freeing the domainlist is the caller's responsibility.
*/
char	*new_domainlist(server_rec *s)
{
	char			*dp, *domainlist = NULL;
	char			*host;
	int 			hostlen;
	int 			i, dlsize = 1;
	kssl_srvr_cfg	*scfg;
	array_header	*ar_scheme = NULL, *ar_https = NULL;

	if ((scfg =	(kssl_srvr_cfg *)
			ap_get_module_config(s->module_config, &kssl_module)) == NULL)  {
		return NULL;
		}

	if (scfg->k_rescheme  &&  (ar_scheme = ap_table_elts(scfg->k_rescheme)))  {
		for (i=0; i < ar_scheme->nelts; i++)  {
			dlsize += strlen(((table_entry *) ar_scheme->elts)[i].key) + 1;
			}
		}
	if (scfg->k_https	  &&  (ar_https  = ap_table_elts(scfg->k_https)))  {
		for (i=0; i < ar_https->nelts; i++)  {
			dlsize += strlen(((table_entry *) ar_https->elts)[i].key) + 1;
			}
		}

	if (!(domainlist = (char *) calloc(1, dlsize)))  {
		return NULL;
		}
	dp = domainlist;

	for (i=0; ar_scheme  &&  i < ar_scheme->nelts; i++)  {
		hostinfo(((table_entry *) ar_scheme->elts)[i].key, &host, &hostlen);
		strncat(dp, host, hostlen);
		dp += strlen(dp) + 1;
		}
	for (i=0; ar_https   &&  i < ar_https->nelts;  i++)  {
		hostinfo(((table_entry *) ar_https->elts)[i].key,  &host, &hostlen);
		strncat(dp, host, hostlen);
		dp += strlen(dp) + 1;
		}

	return domainlist;
}


/*	Read NS cookie file(s) and update entry for kssl proxy.
**	Returns number of files successfully updated.
*/
int 	update_ns_cookie_file(server_rec *s, Cookie *cookie)
{
	char 	*cfpath = NULL, *cfdata = NULL, *dlist = NULL;
	char	*nonce = (cookie)? cookie->nonce: NULL;
	int 	rc, numfiles = 0;

	if (!nonce)  return 0;

	if ((rc = read_file(".netscape/cookies", &cfpath, &cfdata)))  {
		ap_log_error(APLOG_MARK, APLOG_WARNING, s,
			"Unable to read[%d] cookie file %s, skipping\n",
			rc, cfpath? cfpath: "???");
		}
	else if ((dlist = new_domainlist(s)) == NULL)  {
		ap_log_error(APLOG_MARK, APLOG_WARNING, s,
			"Unable to alloc domain list for %s, skipping\n", cfpath);
		}
	else if (!(numfiles = write_ns_cookie_file(dlist,nonce, cfpath,cfdata)))  {
		ap_log_error(APLOG_MARK, APLOG_WARNING, s,
			"Unable to write cookie file %s, skipping\n", cfpath? cfpath:"???");
		}

	if (cfpath)  	free(cfpath);
	if (cfdata)  	free(cfdata);
	if (dlist)		free(dlist);

	return numfiles;
}


/*	Read NS preferences.js file(s) and update entries for kssl proxy.
**	Entries are of form:  user_pref("key", "value");
**	for keys in:
**		network.proxy.http, network.proxy.http_port,
**		network.proxy.ssl,  network.proxy.ssl_port,
**		intl.accept_languages
**	Returns number of files successfully updated.
*/
int 	update_ns_prefs_file(server_rec *s, Cookie *cookie)
{
	char 	*cfpath = NULL, *cfdata = NULL;
	char	*nonce = (cookie)? cookie->nonce: NULL;
	int 	rc, numfiles;

	if (!nonce)  return 0;

	if ((rc = read_file(".netscape/preferences.js", &cfpath, &cfdata)))  {
		ap_log_error(APLOG_MARK, APLOG_WARNING, s,
			"Unable to read[%d] prefs file %s, skipping\n",
			rc, cfpath? cfpath: "???");
		}
	else if (!(numfiles = write_ns_prefs_file(s->port,nonce, cfpath,cfdata))) {
		ap_log_error(APLOG_MARK, APLOG_WARNING, s,
			"Unable to write prefs file %s, skipping\n", cfpath? cfpath: "???");
		}

	if (cfpath)  free(cfpath);
	if (cfdata)  free(cfdata);

	return numfiles;
}


/*	Return 1 if given url is in k_https table, 0 otherwise.
*/
int 	url_in_kssl_list(server_rec *s, const char *url)
{
	kssl_srvr_cfg	*scfg;
	const char		*val = NULL;

	if ((scfg =	(kssl_srvr_cfg *)
		ap_get_module_config(s->module_config, &kssl_module)) == NULL) return 0;

	if ((val = ap_table_get(scfg->k_https, url)))  return 1;
	return 0;
}


/*
 *  Per-module initialization
 */
void	kssl_init_Module(server_rec *s, pool *p)
{
	ap_add_version_component(KSSL_VERSION);
}


/*	Return a keytab value.
**	If cfg->k_keytab is undefined, allocate and define a new value as:
**		keytab = strcat("/etc/krb5.keytab.", k_service)
*/
char	*kssl_get_keytab(pool *p, kssl_srvr_cfg *cfg)
{
	char	*keytab = NULL;

	if (! cfg  ||  ! cfg->k_service)  return NULL;

	if (cfg->k_keytab)  return cfg->k_keytab;

	if ((keytab = (char *)
			ap_pcalloc(p, strlen(KSSL_KTAB_PREFIX)+strlen(cfg->k_service)+1))
			!= NULL)  {
		(void) strcpy(keytab, KSSL_KTAB_PREFIX);
		(void) strcat(keytab, cfg->k_service);
		}

	return keytab;
}


/*	ap_table_do() callback to print keys.
**	Keys are KRB5 users, principals, or realms.  Vals are "t".
*/
static int	kdtab_show(void *data, const char *key, const char *val)
{
	request_rec	*r = (request_rec *) data;

	if (r)  {
		ap_rprintf(r, " %s\n", key);
		}
	else  {
		printf(" %s\n", key);
		}

	return TRUE;
}

void	*kssl_show_srvr_cfg(char *tag, kssl_srvr_cfg *cfg)
{
	if (tag  &&  *tag)  printf("%s ", tag);
	printf("kssl_srvr_cfg %p =\n", cfg);
	if (! cfg)  return;
	if (cfg->k_service)	printf("\tk_service = %s\n", cfg->k_service);
	if (cfg->k_keytab)	printf("\tk_keytab  = %s\n", cfg->k_keytab);
	if (cfg->k_rescheme)  {
		printf("\tk_rescheme= ");
		ap_table_do(kdtab_show, NULL, cfg->k_rescheme, NULL);
		printf("\n");
		}
	if (cfg->k_https)  {
		printf("\tk_https   = ");
		ap_table_do(kdtab_show, NULL, cfg->k_https, NULL);
		printf("\n");
		}

	if (cookie)  {
		printf("\tcookie = %p\n", cookie);
		printf("\tnonce [%d] = %s\n",	cookie->nonce_len / 2,
										cookie->nonce? cookie->nonce: "<NULL>");
		printf("\ttries/max = %d / %d\n", cookie->tries, cookie->max_tries);
		printf("\tns/ms_auth= %2X/%2X\n", cookie->ns_auths, cookie->ms_auths);
		}
	return;
}


void	*kssl_show_dir_cfg(char *tag, kssl_dir_cfg *cfg)
{
	if (tag  &&  *tag)  printf("%s ", tag);
	printf("kssl_dir_cfg %p =\n", cfg);
	if (cfg->k_princ)  {
		printf("\tk_princ =");
		ap_table_do(kdtab_show, NULL, cfg->k_princ, NULL);
		printf("\n");
		}
	if (cfg->k_user)   {
		printf("\tk_user  =");
		ap_table_do(kdtab_show, NULL, cfg->k_user, NULL);
		printf("\n");
		}
	if (cfg->k_realm)  {
		printf("\tk_realm =");
		ap_table_do(kdtab_show, NULL, cfg->k_realm, NULL);
		printf("\n");
		}
}


/*	Return all table keys in one string.
*/
char	*kdtab_keys(table *t)
{
	array_header	*a;
	int 			i, strsize;
	char 			*str;

	if (t == NULL  ||  (a = ap_table_elts(t)) == NULL)  return NULL;

	strsize = a->nelts;
	for (i=0; i < a->nelts; i++)  {
		strsize += strlen(((table_entry *) a->elts)[i].key);
		}

	if ((str = (char *) ap_pcalloc(a->pool, strsize+1)) == NULL)  return NULL;

	for (i=0; i < a->nelts; i++)  {
		if (i)  (void) strcat(str, " ");
		(void) strcat(str, ((table_entry *) a->elts)[i].key);
		}

	return str;
}


void	*kssl_create_srvr_cfg(pool *p, server_rec *s)
{
	kssl_srvr_cfg	*cfg = (kssl_srvr_cfg *)
							ap_pcalloc(p, sizeof(kssl_srvr_cfg));
	
	if (cfg)  {
		cfg->k_service	= (char *) ap_pstrdup(p, KSSL_SVC_DEFAULT);
		cfg->k_keytab	= '\0'; 	/* NULL ==> /etc/krb5.keytab.<service>	*/
		cfg->k_rescheme	= ap_make_table(p, 1);
		cfg->k_https	= ap_make_table(p, 1);
		cfg->k_auths	= ap_make_table(p, 1);
		}

	return (void *) cfg;
}

void	*kssl_create_dir_cfg(pool *p, char *path)
{
	kssl_dir_cfg	*cfg = (kssl_dir_cfg *) ap_pcalloc(p, sizeof(kssl_dir_cfg));
	
	if (cfg)  {
		cfg->k_realm	= ap_make_table(p, 1);
		cfg->k_princ	= ap_make_table(p, 1);
		cfg->k_user 	= ap_make_table(p, 1);
		}

	return (void *) cfg;
}


void *kssl_merge_srvr_cfg(pool *p, void *base, void *new)
{
	kssl_srvr_cfg	*merged	= (kssl_srvr_cfg *)
										ap_pcalloc(p, sizeof(kssl_srvr_cfg));
	kssl_srvr_cfg	*parent	= (kssl_srvr_cfg *) base;
	kssl_srvr_cfg	*child 	= (kssl_srvr_cfg *) new;

	merged->k_service = (child->k_service)? child->k_service: parent->k_service;
	merged->k_keytab  = (child->k_keytab )? child->k_keytab : parent->k_keytab;
	merged->k_rescheme=(child->k_rescheme)?child->k_rescheme:parent->k_rescheme;
	merged->k_https   = (child->k_https  )? child->k_https  : parent->k_https;
	merged->k_auths   = (child->k_auths  )? child->k_auths  : parent->k_auths;

	return (void *) merged;
}

/*	Note that tables are replaced rather than merged.  For example, in:
**		<Location /x>		KRB5User fred barney		</Location>
**		<Location /x/y/z>	KRB5User pebbles			</Location>
**	Users fred and barney have access to directories /x and /x/y, 
**	but only pebbles can get into /x/y/z.
*/
void *kssl_merge_dir_cfg(pool *p, void *base, void *new)
{
	kssl_dir_cfg	*merged	= (kssl_dir_cfg *)
											ap_pcalloc(p, sizeof(kssl_dir_cfg));
	kssl_dir_cfg	*parent	= (kssl_dir_cfg *) base;
	kssl_dir_cfg	*child	= (kssl_dir_cfg *) new;

	merged->k_realm   = (child->k_realm  )? child->k_realm  : parent->k_realm;
	merged->k_princ   = (child->k_princ  )? child->k_princ  : parent->k_princ;
	merged->k_user    = (child->k_user   )? child->k_user   : parent->k_user ;

	return (void *) merged;
}


/*	service directive handler
*/
const char	*ks_directive(cmd_parms *parms, void *mconfig, const char *val)
{
	kssl_dir_cfg 	*dcfg = (kssl_dir_cfg *) mconfig;
	kssl_srvr_cfg 	*scfg = (kssl_srvr_cfg *)
			ap_get_module_config(parms->server->module_config, &kssl_module);

	switch ((int) parms->info)  {
		case	KSSL_SVC_N:
					scfg->k_service	= (char *) ap_pstrdup(parms->pool, val);
					break;
		case	KSSL_KTAB_N:
					scfg->k_keytab	= (char *) ap_pstrdup(parms->pool, val);
					break;
		case	KSSL_EDT_N:
					if (!cookie)  cookie = new_cookie();
					ap_table_set(scfg->k_rescheme,  val, "t");
					break;
		case	KSSL_SSL_N:
					if (!cookie)  cookie = new_cookie();
					ap_table_set(scfg->k_https,  val, "t");
					break;
		case	KSSL_NONC_N:
					if (!cookie)  cookie = new_cookie();
					if (cookie)   cookie->nonce = new_nonce(val);
					break;
		case	KSSL_NONL_N:
					if (!cookie)  cookie = new_cookie();
					if (cookie)   cookie->nonce_len = atoi(val);
					break;
		case	KSSL_MAXT_N:
					if (!cookie)  cookie = new_cookie();
					if (cookie)   cookie->max_tries = atoi(val);
					break;
		case	KSSL_AUTH_N:
					if (!cookie)  cookie = new_cookie();
					if (cookie)   parse_auth_method(cookie, val);
					ap_table_set(scfg->k_auths,  val, "t");
					break;
		default:	break;
		}

	return NULL;
}

/*	directory directive handler
*/
const char	*kd_directive(cmd_parms *parms, void *mconfig, const char *val)
{
	kssl_dir_cfg 	*cfg = (kssl_dir_cfg *) mconfig;

	switch ((int) parms->info)  {
		case	KSSL_RLM_N:	ap_table_set(cfg->k_realm, val, "t");	break;
		case	KSSL_PRC_N:	ap_table_set(cfg->k_princ, val, "t");	break;
		case	KSSL_USR_N:	ap_table_set(cfg->k_user,  val, "t");	break;
		default:			break;
		}

	return NULL;
}


/*	define module directives
*/
static const command_rec	kssl_cmds[] =
{
	{ "KRB5Service",	ks_directive, (void *) KSSL_SVC_N,
		RSRC_CONF, TAKE1,
		"a Kerberos service name; default is \"kssl\""	},
	{ "KRB5KeytabFile",	ks_directive, (void *) KSSL_KTAB_N,
		RSRC_CONF, TAKE1,
		"a Kerberos keytab file name; default is \"/etc/krb5.keytab\""	},
	{ "KRB5ProxyRescheme",	ks_directive, (void *) KSSL_EDT_N,
		RSRC_CONF, ITERATE,
		"URLs to edit http -> https"	},
	{ "KRB5ProxyHttps",	ks_directive, (void *) KSSL_SSL_N,
		RSRC_CONF, ITERATE,
		"URLs of permitted https sites"	},
	{ "KRB5ProxyCookie",	ks_directive, (void *) KSSL_NONC_N,
		RSRC_CONF, TAKE1,
		"a user specified cookie, required for access to proxy port"	},
	{ "KRB5ProxyCookieLength",	ks_directive, (void *) KSSL_NONL_N,
		RSRC_CONF, TAKE1,
		"length for random cookie, required for access to proxy port"	},
	{ "KRB5ProxyMaxTries",	ks_directive, (void *) KSSL_MAXT_N,
		RSRC_CONF, TAKE1,
		"maximum number of proxy port access failures"	},
	{ "KRB5ProxyAuthMethod",	ks_directive, (void *) KSSL_AUTH_N,
		RSRC_CONF, ITERATE,
		"list of proxy port authentication methods (NS:Cookie, NS:Language)" },

	{ "KRB5Realm",	kd_directive, (void *) KSSL_RLM_N,
		RSRC_CONF | OR_AUTHCFG, ITERATE,
		"the Kerberos REALM to be required for access" },
	{ "KRB5Principal",	kd_directive, (void *) KSSL_PRC_N,
		RSRC_CONF | OR_AUTHCFG, ITERATE,
		"the Kerberos user principal to be required for access" },
	{ "KRB5Principle",	kd_directive, (void *) KSSL_PRC_N,
		RSRC_CONF | OR_AUTHCFG, ITERATE,
		"the Kerberos user principal to be required for access" },
	{ "KRB5User",	kd_directive, (void *) KSSL_USR_N,
		RSRC_CONF | OR_AUTHCFG, ITERATE,
		"the Kerberos username to be required for access" },
	{ NULL }
};


/*	Copy server config directives to SSL *s.  
**	This needs to be done after creation of SSL *s, but before the call to
**	SSL_accept(ssl) (in mod_ssl ssl_engine_kernel.c).
**	Returns 0 if happy, non-zero otherwise.
*/
int 	kssl_apply_srvr_cfg(conn_rec *conn, SSL *s)
{
	kssl_srvr_cfg	*scfg;

	if (conn == NULL  ||  s == NULL)  return 1;

	if ((scfg =	(kssl_srvr_cfg *)
			ap_get_module_config(conn->server->module_config, &kssl_module))
			== NULL)  return 1;

	if (s->kssl_ctx == NULL && (s->kssl_ctx = kssl_ctx_new()) == NULL) return 1;

	if (scfg->k_service)  {
		s->kssl_ctx->service_name = (char *)
									ap_pstrdup(conn->pool, scfg->k_service);
		s->kssl_ctx->keytab_file = kssl_get_keytab(conn->pool, scfg);
		}
	else  {
		return 1;
		}

	return 0;
}


/*	Child process initialization handler.
**	This is strictly for (kerberized local) proxy usage.
**	The proxy is going to write a nonce value into the user's
**	$HOME/.netscape cookies and/or prefs files which will be used
**	to control access to the proxy's listening port.
**	The nonce depends on values read from server config, but has
**	to be written before any requests come in.  
*/
void 	kssl_child_init(server_rec *s, pool *p)
{
	kssl_srvr_cfg	*scfg;
	char 			*cp;
	unsigned char	r;
	int 			i = 0;
	array_header	*a;

	if (!cookie  ||
		(scfg = (kssl_srvr_cfg *)
				ap_get_module_config(s->module_config, &kssl_module))
				== NULL)  return;

	if (scfg->k_rescheme  && (a=ap_table_elts(scfg->k_rescheme))) i+= a->nelts;
	if (scfg->k_https     && (a=ap_table_elts(scfg->k_https   ))) i+= a->nelts;
	
	if (! cookie->nonce  &&  cookie->nonce_len)  {
		if ((cookie->nonce = (char *) calloc(2, cookie->nonce_len+1)) == NULL) {
			return;
			}
		cp = cookie->nonce;
		for (i=0; i < cookie->nonce_len; i++)  {
			r = randbyte();
			*cp = (r & 0x0F)  +  0x30;			/* byte to ascii conversion	*/
			if (*cp > 0x39)  *cp += 7;
			cp++;
			*cp = ((r >> 4) & 0x0F)  +  0x30;
			if (*cp > 0x39)  *cp += 7;
			cp++;
			}
		cookie->nonce_len *= 2;
		}

	if (cookie->ns_auths & AUTH_COOK)  (void) update_ns_cookie_file(s, cookie);
	if (cookie->ns_auths & AUTH_LANG)  (void) update_ns_prefs_file(s, cookie);

	/*	Unimplemented...
	if (cookie->ms_auths & AUTH_COOK)  (void) update_ms_cookie_file(s, cookie);
	if (cookie->ms_auths & AUTH_LANG)  (void) update_ms_prefs_file(s, cookie);
	*/

	return;
}


/*	Child process exit handler.
**	This is strictly for (kerberized local) proxy usage.
**	Scavenge old nonce values from $HOME/.netscape cookies and/or prefs files
**	so they won't come back to haunt us later.
*/
void 	kssl_child_exit(server_rec *s, pool *p)
{
	int 	bp = s->port;

	if (!cookie)  return;

	/* s->port=0 flags update_ns_prefs_file() to restore previous saved values
	*/
	s->port = 0;
	(void) update_ns_prefs_file(s, cookie);
	s->port = bp;

	return;
}


/*	Tweaked for use in mitm proxy; works on byte stream vs header table.
*/
/*	Check for, and edit out, proxy port access cookie in http headers.
**	Here str is the value of a Cookie: or Accept-Language: header and the
**	cookie nonce value is expected to follow key ("ksslproxy=" or "ksslproxy-").
**	Returns the nonce value if found, NULL otherwise.
**	Also returns a censored copy of the header value string.
*/
char 	*eat_cookie(pool *p, char *str, char *key, char **newstr)
{
	char 	*found, *nonce, *extra;

	if (!str  ||  !*str  ||  !key  || !*key)  return NULL;

	if (!(found = strstr(str, key)))  return NULL;
	if (!strlen(nonce = found + strlen(key)))  return NULL;

	*newstr = ap_pstrndup(p, str, found - str);
	if ((found = strrchr(*newstr, ','))  &&  strlen(found) == 2  &&
		isspace(*(found+1)))  *found = '\0';
	if ((found = strpbrk(nonce, " \t\r\n")))  {
		extra = (*found == '\n')? "\n": " ";		/*  for multiline  */
		*found++ = '\0';
		if (*found)  *newstr = ap_pstrcat(p, *newstr, extra, found, NULL);
		}

	return nonce;
}


/*	Check for nonce in Cookie: or Accept-Language: headers.
**	Returns FORBIDDEN if proxy port access control is in effect
**	and nonce is missing or incorrect.
**	Returns DECLINED otherwise.
**	Destroys nonce in outgoing headers in any case.
*/
int 	kssl_check_headers(request_rec *r)
{
	kssl_srvr_cfg	*scfg;
	array_header	*a;
	int 			i;
	char			*header, *value, *new_val, *hd_nonce = "";
	int 			rc = FORBIDDEN;

	if ((scfg =	(kssl_srvr_cfg *)
			ap_get_module_config(r->server->module_config, &kssl_module))
			== NULL)  return DECLINED;

	if (r->headers_in  &&  (a = ap_table_elts(r->headers_in)))  {
		for (i=0; i < a->nelts; i++)  {
			header = ((table_entry *) a->elts)[i].key;
			value  = ((table_entry *) a->elts)[i].val;

			if (strcasecmp("Accept-Language", header) == 0)  {
				if (hd_nonce = eat_cookie(r->pool, value,
												COOKIE_TAG "-", &new_val))  {
					((table_entry *) a->elts)[i].val = new_val;
					if (cookie  &&  (cookie->ns_auths & AUTH_LANG)  &&
						cookie->nonce  &&  !strcmp(hd_nonce, cookie->nonce))  {
						rc = DECLINED;
						}
					}
				}
			if (strcasecmp("Cookie", header) == 0)  {
				if (hd_nonce = eat_cookie(r->pool, value,
												COOKIE_TAG "=", &new_val))  {
					((table_entry *) a->elts)[i].val = new_val;
					if (cookie  &&  (cookie->ns_auths & AUTH_COOK)  &&
						cookie->nonce  &&  !strcmp(hd_nonce, cookie->nonce))  {
						rc = DECLINED;
						}
					}
				}
			}
		}

	if (!cookie  ||  !cookie->nonce_len)  rc = DECLINED;
	if (r->method_number == M_CONNECT)  rc = DECLINED;

	if (cookie  &&  rc == FORBIDDEN)  {
		cookie->tries++;
		if (cookie->tries > cookie->max_tries)  {
			ap_log_error(APLOG_MARK, APLOG_NOERRNO | APLOG_EMERG, r->server, 
				"Exceeded max (%d) bad connection attempts to proxy", 
				cookie->max_tries);
			exit(1);
			}
		}

	return rc;
}


#ifdef KSSL_KLP

/*	Error response for bad/missing auth cookie
*/
const char	*badcookie_response =
		"HTTP/1.0 403 Access to proxy denied\n"
		"Proxy-agent: apache/kssl\n"
		"Content-type: text/html\n"
		"Content-length: CONTLEN\n"
		"\n"
		"<HTML><Head><Title> Access to proxy denied</Title></Head>\n\n"
		"<Body><H1> 403 Access to proxy denied</H1>\n<P>\n"
		"Access to the local proxy was rejected because \n"
		"your web browser did not provide the required cookie. \n"
		"Cookies are provided in the <Code> Cookie:</Code> or \n"
		"<Code> Accept-Language:</Code> HTTP headers. \n"
		"<P>\nPossible remedies:\n<OL>\n"
		"<LI>  You must run the kerberized local proxy\n"
			   "<Strong>before</Strong> you start up your web browser\n"
		"<!--"
		"<LI>  You must enable cookies in your web browser\n"
		"--> "
		"</OL>\n"
		"</Body></HTML>\n" ;


/*	Search for cookie in local channel.
**	This duplicates cookie checking code for use in mitm proxy,
**	where cookies are found in a byte stream, rather than in 
**	nicely parsed header_in tables.
**
**	Return 1:OK if good cookie found, or otherwise OK.
**	Otherwise return 0:Fail and
**		o	Write error message to local channel
**		o	Exit proxy if failed connection attempt limit exceeded.
*/
int 	got_cookie(struct pool *p, Channel *local, Channel *remote)
{
	Channel		badcookie_channel;
	const int	badcookie_len = strlen(badcookie_response);
	const int	com1 = strstr(badcookie_response, "<!--")
					- badcookie_response;
	const int	com2 = strstr(badcookie_response, "--> ")
					- badcookie_response;
	char		*cp;
	int 		have_cookie = 0;
	char		*str, *newstr;
	char		*hd_nonce1 = "", *hd_nonce2 = "";
	int 		len;

	if (!(str = ap_pstrndup(p, local->buff, local->ptr)))  return 0;

	/*	Always destroy cookie before sending data to remote side.
	*/
	hd_nonce1 = eat_cookie(p, str, COOKIE_TAG "-", &newstr);
	if (newstr)  str = newstr;
	hd_nonce2 = eat_cookie(p, str, COOKIE_TAG "=", &newstr);
	if ((newstr = (newstr)? newstr: str))  {
		local->ptr = strlen(newstr);
		strncpy(local->buff, newstr, local->ptr);
		}
	
	if (!cookie  ||  !cookie->nonce_len)  return 1;
	if (!cookie->nonce)  return 0;
	
	if ((cookie->ns_auths & AUTH_LANG)  &&  hd_nonce1  &&
		!strcmp(hd_nonce1, cookie->nonce))  have_cookie = 1;
	if ((cookie->ns_auths & AUTH_COOK)  &&  hd_nonce2  &&
		!strcmp(hd_nonce2, cookie->nonce))  have_cookie = 1;

	if (have_cookie)  return have_cookie;

	if (badcookie_len + 1 > BUFFSIZE)  {
		fprintf(stderr, "Programming botch: "
						"badcookie_response exceeds BUFFSIZE (%d)\n", BUFFSIZE);
		exit(1);
		}
	badcookie_channel.ptr = badcookie_len;
	strcpy(badcookie_channel.buff, badcookie_response);
	if ((cp = strstr(badcookie_channel.buff, "CONTLEN")))  {
		snprintf(cp, 7, "%-7d", strlen(cp+9));
		if (strlen(cp) < 10)  cp[strlen(cp)] = '\n';
		}
	if (cookie->ns_auths & AUTH_COOK)  {
		strncpy(&badcookie_channel.buff[com1], "    ", 4);
		strncpy(&badcookie_channel.buff[com2], "    ", 4);
		}
	(void) wr_ssl_ch(&badcookie_channel, local, &len);

	if (++cookie->tries > cookie->max_tries)  {
		fprintf(stderr, "proxy exiting after %d failed connection attempts\n",
				cookie->tries);
		exit(1);
		}

	return have_cookie;
}

#endif  /* KSSL_KLP */


/*	Check user Kerberos principal against directive restrictions.
**	Copy kssl_ctx->client_princ data to apache table variables.
**	Returns DECLINED if happy, FORBIDDEN otherwise.
**	Currently set as auth handler.
**
*/
int 	kssl_check_user_auth(request_rec *r)
{
	table			*e;
	const char		*val;
	char			*at = NULL;
	char			*princ = NULL, *user = NULL, *realm = NULL;
	SSL 			*s;
	kssl_dir_cfg	*cfg = NULL;
	SSLDirConfigRec	*ssl_dc;
	SSLSrvConfigRec	*ssl_sc;
	unsigned int	bad = 0;

	/*	This is the "check access by host address" auth method of mod_ssl.c.
	**	It needs to be called before this mod_kssl auth method in order to
	**	do cipher renegotiation.  The ssl_hook_Access() routine is now 
	**	being called twice.  It would be preferable to change the order of
	**	module invocation to mod_ssl->mod_kssl; I was unable to do this by
	**	changing the configuration order of --enable- and --activate- module.
	*/
	if (ssl_hook_Access(r) == FORBIDDEN)  return FORBIDDEN;

	if (r == NULL  || 
		r->connection == NULL  ||
		r->connection->client == NULL  ||
		r->connection->client->ctx == NULL)  {
		ap_log_reason("kssl user auth sanity check bounced",r->filename,r);
		ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
		return FORBIDDEN;
		}

	if ((ssl_dc = myDirConfig(r)) == NULL  ||  
		(ssl_sc = mySrvConfig(r->server)) == NULL)  {
		ap_log_reason("kssl user auth sanity_check bounced",r->filename,r);
		ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
		return FORBIDDEN;
		}

	if ((s = ap_ctx_get(r->connection->client->ctx, "ssl")) == NULL  &&
		ssl_dc->bSSLRequired)  {
		ap_log_reason("SSL connection required",r->filename,r);
		ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
		return FORBIDDEN;
		}
	
	/*	VRS 19991201
	**	ssl_dc->bSSLRequired, ssl_sc->bEnabled, and SSL *s are defined 
	**	only when in an SSL VirtualHost.
	**	ssl_dc->bSSLRequired, ssl_sc->bEnabled, and SSL *s all go to zero
	**	when in the regular document root, granting equal access to both
	**	regular and kerberos pages.  If you have a kerberos dir located
	**	under the regular document root make sure to include an
	**	SSLRequireSSL directive in the regular document root's path to
	**	the kerberos pages.
	*/
	if (!ssl_sc->bEnabled  ||  s == NULL)  return DECLINED;

	if (r->per_dir_config == NULL  ||
		(cfg = (kssl_dir_cfg *)
				ap_get_module_config(r->per_dir_config, &kssl_module)) 
				== NULL)  {
		ap_log_reason("kssl user_auth sanity check bounced",r->filename,r);
		ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
		return FORBIDDEN;
		}

	if (ap_is_empty_table(cfg->k_realm)  &&
		ap_is_empty_table(cfg->k_user)   &&
		ap_is_empty_table(cfg->k_princ))  return DECLINED;

	if (s->kssl_ctx == NULL  ||
		(princ = s->kssl_ctx->client_princ) == NULL  ||  
		(user = (char *) ap_pstrdup(r->pool, princ)) == NULL)  {
		ap_log_reason("kssl_user auth sanity check bounced",r->filename,r);
		ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
		return FORBIDDEN;
		}
	if ((e = r->subprocess_env) == NULL)  {
		ap_log_reason("kssl_user_auth sanity check bounced",r->filename,r);
		ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
		return FORBIDDEN;
		}

	/*	Push variables into r->subprocess_env table,
	**	in hopes they will be available to cgi-bin/ scripts.
	*/
	if ((at = strchr(user, '@')) != NULL)  {
		*at = '\0';
		realm = at + 1;
		ap_table_set(e, "KRB5_REALM", realm);
		ap_table_set(e, "KRB5_USER",  user);
		ap_table_set(e, "KRB5_PRINC", princ);
		}
	else  {
		/*  Log malformed kssl_ctx->user: must be user@Realm  */
		ssl_log(r->server, SSL_LOG_INFO, "kssl malformed user principal: %s",
				princ);
		ap_log_reason("kssl user auth malformed user principal",r->filename,r);
		ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
		return FORBIDDEN;
		}

	if (cfg->k_princ &&  !ap_is_empty_table(cfg->k_princ)  &&
		(val = ap_table_get(cfg->k_princ, princ)) == NULL) bad |= KSSL_BAD_PRC;
	if (cfg->k_realm &&  !ap_is_empty_table(cfg->k_realm)  &&
		(val = ap_table_get(cfg->k_realm, realm)) == NULL) bad |= KSSL_BAD_RLM;
	if (cfg->k_user  &&  !ap_is_empty_table(cfg->k_user)   &&
		(val = ap_table_get(cfg->k_user, user)) == NULL)   bad |= KSSL_BAD_USR;

	if (bad)  {
		/*	Log reason for failure: bad  */
		if (bad & KSSL_BAD_RLM)  ssl_log(r->server, SSL_LOG_INFO,
					"kssl access denied: user %s not in KRB5Realm %s", princ,
													kdtab_keys(cfg->k_realm));
		if (bad & KSSL_BAD_PRC)  ssl_log(r->server, SSL_LOG_INFO,
					"kssl access denied: user %s not in KRB5Principal %s",princ,
													kdtab_keys(cfg->k_princ));
		if (bad & KSSL_BAD_USR)  ssl_log(r->server, SSL_LOG_INFO,
					"kssl access denied: user %s not in KRB5User %s", princ,
													kdtab_keys(cfg->k_user));

		ap_log_reason("kssl access denied",r->filename,r);
		ap_table_setn(r->notes, "ssl-access-forbidden", (void *)1);
		return FORBIDDEN;
		}

	return DECLINED;
}


/*	URI translation method (conditionally edit http: -> https:)  VRS 20000121
**	for  KRB5ProxyRescheme http://ksslhost1:port/...
**
**	Can demo Netscape/MS-IE --> kssl proxy --> kssl web server as follows:
**	Set web browser to proxy _regular_ http.  Proxying https would put
**	the browsers into CONNECT mode.  Request kerberos auth'd web page
**	as http: rather than https: method.
**
**	This s/http/https/ translation happens very early in the apache
**	module sequence, prior to proxy_detect().  The proxy will see
**	"GET https://..." and all is well at this stage.  BUT: any
**	embedded https: links will fail (CONNECT) or would need to be
**	content-handler edited into http: links.
**	This doesn't seem like a good solution.
*/
int 	kssl_xlate_proxy(request_rec *r)
{
	kssl_srvr_cfg	*scfg = (kssl_srvr_cfg *) 
					ap_get_module_config(r->server->module_config,&kssl_module);
	array_header	*a;
	int 			i;
	char			*xlate_uri;

	if (!scfg  ||  !scfg->k_rescheme  ||
		(a = ap_table_elts(scfg->k_rescheme)) == NULL)  return DECLINED;

	for (i=0; i < a->nelts; i++)  {
		if (strncmp(r->unparsed_uri, ((table_entry *) a->elts)[i].key,
			strlen(((table_entry *) a->elts)[i].key)) == 0)  {
			xlate_uri = ap_pstrcat(r->pool, "https", &r->unparsed_uri[4], NULL);
			r->uri = xlate_uri;
			r->unparsed_uri= xlate_uri;
			}
		}

	return DECLINED;
}


/*	Spew status for debugging.
**	<Location /foo/status/>
**		SetHandler kssl-status-handler
*/
int 	kssl_status_handler(request_rec *r)
{
	const char		*hostname;
	char			*keytab; 
	char			*var, *val;
	table			*e;
	kssl_srvr_cfg	*scfg = (kssl_srvr_cfg *)
						ap_get_module_config(r->server->module_config,
																&kssl_module);
	kssl_dir_cfg	*dcfg = (kssl_dir_cfg *)
						ap_get_module_config(r->per_dir_config, &kssl_module);

	keytab = kssl_get_keytab(r->pool, scfg);

	r->content_type = "text/html";
	ap_send_http_header(r);
	hostname = ap_get_remote_host(r->connection,r->per_dir_config, REMOTE_NAME);

	ap_rputs("<HTML>\n<Head> <Title> kssl debug</Title>\n</Head>\n\n", r);
	ap_rputs("<Body>\n", r);
	ap_rprintf(r, "<H1> kssl request status:</H1>\n");
	ap_rprintf(r, "Request from: <Strong> %s</Strong>\n", hostname);
	ap_rputs("<P>\n", r);

	ap_rputs("<P> <H2> Current Config Data</H2>\n", r);
	ap_rprintf(r, "<Strong> k_service</Strong> = %s<br>\n", scfg->k_service);
	ap_rprintf(r, "<Strong> k_keytab</Strong> = %s<br>\n", keytab);
	ap_rprintf(r, "<Strong> k_realm</Strong> = ");
		ap_table_do(kdtab_show, r, dcfg->k_realm, NULL);
		ap_rprintf(r, "<br>\n");
	ap_rprintf(r, "<Strong> k_user</Strong> = ");
		ap_table_do(kdtab_show, r, dcfg->k_user, NULL);
		ap_rprintf(r, "<br>\n");
	ap_rprintf(r, "<Strong> k_princ</Strong> = ");
		ap_table_do(kdtab_show, r, dcfg->k_princ, NULL);
		ap_rprintf(r, "<br>\n");

	ap_rputs("<P> <H2> ssl_var_lookup Data</H2>\n", r);
	var = "SSL_CIPHER";
	val = ssl_var_lookup(r->pool, r->server, r->connection, r, var);
	ap_rprintf(r, "<Strong> %s</Strong> = %s<br>\n", var, val);

	ap_rputs("<P> <H2> subprocess_env table Data</H2>\n", r);
	if ((e = r->subprocess_env) != NULL)  {
		var = "KRB5_PRINC";
		val = (char *) ap_table_get(e, var);
		ap_rprintf(r, "<Strong> %s</Strong> = %s<br>\n", var, val);
		var = "KRB5_REALM";
		val = (char *) ap_table_get(e, var);
		ap_rprintf(r, "<Strong> %s</Strong> = %s<br>\n", var, val);
		var = "KRB5_USER";
		val = (char *) ap_table_get(e, var);
		ap_rprintf(r, "<Strong> %s</Strong> = %s<br>\n", var, val);
		}

	ap_rputs("<P>\n",r );
	ap_rputs("\n</Body>\n\n</HTML>\n", r);

	return OK;
}

/* define name of content handler to apache
*/
static handler_rec	kssl_handlers[] = 
{
	{"kssl-status-handler",	kssl_status_handler},
	NULL
};


/*	define phases of transaction we handle to apache
*/
module MODULE_VAR_EXPORT	kssl_module =
{
	STANDARD_MODULE_STUFF,
	kssl_init_Module,			/* module initializer */
	kssl_create_dir_cfg,	 	/* per-directory config creator */
	kssl_merge_dir_cfg,			/* dir config merger */
	kssl_create_srvr_cfg,		/* server config creator */
	kssl_merge_srvr_cfg,			/* server config merger */
	kssl_cmds,					/* command table */
	kssl_handlers,				/* [9]	content handlers	*/
	NULL,			/* [2]	uri to filename translation */
	NULL,						/* [5]	check/validate user id */
	NULL,						/* [6]	check user id is valid *here* */
	kssl_check_user_auth,		/* [4]	check access by host address */
	NULL,						/* [7]	mime type checker/converter */
	NULL,						/* [8]	fixups */
	NULL,						/* [10]	logger */
	kssl_check_headers,			/* [3]	header parser */
	kssl_child_init,			/* process initialization */
	kssl_child_exit,			/* process exit/cleanup */
	kssl_xlate_proxy,			/* [1]	post read request handling */
};


#endif	/* NO_KRB5 */

